function factor = snapback(A)
% factor = snapback(A)
% 
% Computes the snap-back factorization of a banded symmetric matrix A,
% according to the paper by Irony and Toledo, "The snap-back pivoting
% method for symmetric banded matrices".
%
% THIS CODE IS A DEMO OF THE ALGORITHM, NOT PRODUCTION SOFTWARE.
% In particular, it is meant to illustrate the algorithm, not to be
% efficient. The factor object is an array of structures, each describing
% the elimination type for that step and the transformation matrices. The
% matrices correspond exactly to the notation in the paper.
% 
% To solve Ax=b with the factor, use x = snapback_solve(factor,b).
% 
% To generate a suitable banded matrix for testing, use banded(n,m,l),
% where n is the dimension, m is the half-bandwidth, and l is the number of
% negative eigenvalues.
% 
% We provide a matrix that causes all 3 elimination types to occur; use
% load A.mat -ascii
% to load it, then run factor=snapback(A);

close;
spy(A);
title('Original Matrix');
pause;


alpha = 1/3;

n = size(A,1);
normA = norm(A,1);

%for m = n:-1:0
%	if (max(abs(diag(A,m))) ~= 0)
%		break;
%	end
%end
	
m = half_bandwidth(A);
fprintf(1,'The half-bandwidth of A is m=%d\n',m);

for j=1:n
	  factor(j).type=0;
end;	

j = 1;
while (j < n)
	
	% check for type-1 elimination
	
	if (abs(A(j,j)) > alpha*max(abs(A(j+1:n,j))))
		% type-1 elimination
		factor(j).type = 1;
		
		pivot = A(j,j);
		
		Lj           = eye(n);
		Lj(j+1:n,j)  = A(j+1:n,j)/pivot;
		
		factor(j).L  = Lj;
		
		iLj          = eye(n);
		iLj(j+1:n,j) = -A(j+1:n,j)/pivot;

		A = iLj * A * iLj';
		
		j = j+1;
 
		spy(abs(A) > 1e-12*normA);
		title('Type-1 Elimination; Applied inv(L)');
		pause;
	else
		% we need a type-2 or type-3 elimination
	
		%find local band at row/column j
		limit = (j-1) + max(find( A(j:n,j)));
		
		Y=eye(n);
		for(i=j+1:limit-1)
			iYij=eye(n);
			if(A(i+1,j)~=0 | A(i,j)~=0)
				iYij(i,i)=A(i+1,j)/sqrt(A(i+1,j)^2+A(i,j)^2);
				iYij(i+1,i+1)=iYij(i,i);;
				iYij(i+1,i)=A(i,j)/sqrt(A(i+1,j)^2+A(i,j)^2);
				iYij(i,i+1)=-iYij(i+1,i);
			end;
			A=iYij*A*iYij';
			Y=Y*iYij'; %equivalent to iY=Y*inv(iYij); but cheaper
		end;
		factor(j).Y  = Y;
		
		spy(abs(A) > 1e-12*normA);
		title('Type-2 or 3 Elimination; Applied inv(Y)');
		pause;
		
		%Calculate G
		iG=eye(n);
		iG(j,j)=A(j,j)/sqrt(A(j,j)^2+A(limit,j)^2);
		iG(limit,limit)=iG(j,j);
		iG(limit,j)=-A(limit,j)/sqrt(A(j,j)^2+A(limit,j)^2);
		iG(j,limit)=-iG(limit,j);
		A=iG*A;
		G=iG';
		factor(j).G  = G;
		
		spy(abs(A) > 1e-12*normA);
		title('Type-2 or 3 Elimination; Applied inv(G)');
		pause;
		
		%Calculate U
		U=eye(n);
		limit_U=j-1+max(find( A(j,j:n)));
		U(j,j:limit_U)=A(j,j:limit_U);
		A(j,j+1:limit_U)=zeros(1,limit_U-j);
		A(j,j)=1;
		factor(j).U = U;
		
		spy(abs(A) > 1e-12*normA);
		title(sprintf('Type-2 or 3 Elimination; Applied inv(U)'));
		pause;
		
		if(abs(A(limit,limit))<max(abs(A(limit,j+1:n))))
			%type-2 elimination
			
			factor(j).type=2;
			
			S=eye(n);
			S(limit,limit)=G(j,j);
			A(limit,j+1:limit-1)=A(j+1:limit-1,limit);
			A(limit,limit+1:limit_U)=A(limit+1:limit_U,limit);
			A(limit,limit)=A(limit,limit)/S(limit,limit);
			factor(j).S=S;
			
			j=j+1;
			spy(abs(A) > 1e-12*normA);
			title(sprintf('Type-2 Elimination (for sure now); Applied inv(S)'));
			pause;
			
		else
			%type-3 elimination
			
			factor(j).type=3;
			
			P=eye(n);
			P(j+1:limit,j+1:limit)=diag(ones(limit-j-1,1),1);
			P(j+1,j+1)=0;
			P(limit,j+1)=1;
			A=P'*A*P; % equivalent to A=inv(P)*A*P
			factor(j).P=P;
			
			spy(abs(A) > 1e-12*normA);
			title(sprintf('Continuing with a Type-3 Elimination; Applied inv(P)'));
			pause;
			
			X=eye(n);
			for(i=j+2:limit-1)
				iXij=eye(n);
				if(A(i+1,j+1)~=0 | A(i,j+1)~=0)
					iXij(i,i)=A(i+1,j+1)/sqrt(A(i+1,j+1)^2+A(i,j+1)^2);
					iXij(i+1,i+1)=iXij(i,i);
					iXij(i+1,i)=A(i,j+1)/sqrt(A(i+1,j+1)^2+A(i,j+1)^2);
					iXij(i,i+1)=-iXij(i+1,i);
				end;
				A=iXij*A*iXij';
				X=X*iXij'; %equivalent to iY=Y*inv(iYij); but cheaper
			end;
			factor(j).X=X;
			
			spy(abs(A) > 1e-12*normA);
			title(sprintf('Type-3 Elimination; Applied inv(X)'));
			pause;
			
			K           = eye(n);
			K(j+1,max(limit,j+2):limit_U)  = A(j+1,max(limit,j+2):limit_U)/A(j+1,j+1);
			
			iK          = eye(n);
			iK(j+1,max(limit,j+2):limit_U) = -K(j+1,max(limit,j+2):limit_U);
			
			A = A * iK;
			factor(j).K=K;
			
			K_2           = eye(n);
			K_2(max(limit,j+2):limit_U,j+1)    = A(max(limit,j+2):limit_U,j+1);
			K_2(j+1,j+1)              = A(j+1,j+1);
			
			A(j+1,j+1)=1;
			A(max(limit,j+2):limit_U,j+1)=zeros(limit_U-max(limit,j+2)+1,1);
			factor(j).K_2=K_2;
			
			j=j+2;
			
			spy(abs(A) > 1e-12*normA);
			title(sprintf('Type-3 Elimination; Applied inv(K) and inv(K-hat)'));
			pause;
		end;
		
	end
	
	spy(abs(A) > 1e-12*normA);
	title(sprintf('%d Rows/Columns Eliminated',j-1));
	pause;
	
end

if(factor(n-1).type~=3)
	factor(n).type=1;
	factor(n).L=eye(n);
end;
for j=1:n
	factor(j).D=A(j,j);
end;

	function m = half_bandwidth(A)
		n = size(A,1);
		for m = n:-1:0
			if (max(abs(diag(A,m))) ~= 0)
				break;
			end
		end
	end

end
